
# $Id: config.py,v 1.14 2004/01/05 15:23:53 cran Exp $
#
# Copyright (c) 2002 Sebastian Stark
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Configuration tree for tentakel

Provides a ConfigBase class that is initialized from a configuration file
with items that are derived from the UserDict.UserDict Python class.

Example:

  c = config.ConfigBase()
  c.load("tentakel.conf")
  availableGroups = c.getGroups()

or, if you want all hosts with expanded sublists:

  hosts = c.getExpandedGroup("mygroup")

In the latter case, only "host" objects are returned in a list.

"""

from error import *
import re
from UserDict import UserDict
import os
import pwd
import tempfile
import tpg

PARAMS = {	'ssh_path': "",
		'method': "",
		'user': "",
		'format': ""
	}

METHODS = ['ssh']

class TConf(tpg.Parser):
	r"""

	set lexer = ContextSensitiveLexer

	token keyword	: 'ssh_path|method|user|format'	str ;
	token eq	: '='		str ;
	token word	: '\w+'		str ;
	token vchar	: '""|[^"]'	str ;
	token hitem	: '\+[\w\.:]+'	str ;
	token litem	: '@\w+'	str ;

	separator spaces	: '\s+' ;
	separator comments	: '#.*' ;

	START/e ->			$ e = {"groups": {}, "settings": {}}
			( SETTING/s	$ e["settings"].update(s)
			)*
			( GROUP/g	$ e["groups"][g["name"]] = g
			)*
	;

	SETTING/s ->	'set'		$ s = {}; s = {"user": pwd.getpwuid(os.geteuid())[0]}
			PARAM/<p,v>	$ s[p] = v
	;

	PARAM/<p,v> ->	keyword/p eq
			'"'
			@start (vchar)* @end	$ t = self.extract(start, end)
			'"'			$ v = re.sub('""', '"', t)
	;

	GROUP/g ->      'group'		$ g = ConfigGroup()
			GROUPNAME/n	$ g["name"] = n
			GROUPSPEC/s	$ g.update(s)
			MEMBERS/l	$ g.update(l)
	;

	GROUPNAME/n -> word/n ;

	GROUPSPEC/s ->					$ s = {}
			'\('
				( PARAM/<p,v>		$ s[p] = v
				)?
				( ',' PARAM/<p,v>	$ s[p] = v
				)*
			'\)'
	;

	MEMBERS/l ->			$ l = {"hosts": [], "lists": []}
			( hitem/i	$ l["hosts"].append(i[1:])
			| litem/i	$ l["lists"].append(i[1:])
			)*
	;
        """

class ConfigGroup(UserDict):
	"store group info"

	def __init__(self):
		UserDict.__init__(self)
		self["name"] = ""
		self.update(PARAMS)
		self["hosts"] = []
		self["lists"] = []
	
	def __str__(self):
		l = []
		for param in PARAMS.keys():
			if self[param]:
				l.append('%s="%s"' % (param, re.sub('"', '""', self[param])))
		return "group %s (%s)" % (self["name"], ', '.join(l))

class ConfigBase(UserDict):
	"store all configuration parameters"

	def __init__(self):
		UserDict.__init__(self)
		self.clear()
	
	def clear(self):
		"make configuration empty"
		self["groups"] = {}
		self["settings"] = PARAMS

	def parse(self, txt):
		
		tp = TConf()
		self.update(tp(txt))
	
	def load(self, filename):
		"load configuration from file"
		
		try:
			f = open(filename)
			try:
				self.parse("".join(f.readlines()))
			except tpg.SyntacticError, err:
				warn("in %s: %s" % (filename, err.msg))
		except IOError:
			err("could not read file: %s" % filename)
		f.close()
	
	def dump(self, filename):
		"save configuration to file"
		
		try:
			f = open(filename, "w")
			f.writelines(str(self))
		except IOError:
			err("could not write file: %s" % filename)
		f.close()
	
	def edit(self):
		"interactively edit configuration"

		# TODO: For now, I don't consider this a security hole
		# (symlink attack etc.) because the file is opened in an
		# editor and the user can see it.
		# When Python 2.3 becomes more standard the new tempfile
		# methods are going to be used.
		try:
			tempedit = tempfile.mktemp()
			self.dump(tempedit)
			editor = os.getenv("VISUAL") or os.getenv("EDITOR") or "vi"
			os.spawnvp(os.P_WAIT, editor, [editor, tempedit])
			self.load(tempedit)
		finally:
			os.remove(tempedit)

	def __str__(self):
		"pretty print configuration"

		out = ""
		settings = self["settings"]
		for s_param, s_value in settings.items():
			if s_value:
				out = "%sset %s=\"%s\"\n" % (out, s_param, s_value)
		out = out + "\n"
		groups = self["groups"]
		for groupName, groupObj in groups.items():
			out = out + str(groupObj) + "\n"
			for list in groups[groupName]["lists"]:
				out = out + "\t@" + list + "\n"
			for host in groups[groupName]["hosts"]:
				out = out + "\t+" + host + "\n"
		return out

	def getGroup(self, groupName):
		"Return the group groupName"

		return self["groups"][groupName]
	
	def getGroups(self):
		"Return list of group names"

		return self["groups"].keys()
	
	def getExpandedGroup(self, groupName):
		"Return list of groupName members with sub lists expanded recursively"

		g = self.getGroup(groupName)
		out = g["hosts"]
		for list in g["lists"]:
			try:
				out = out + self.getExpandedGroup(list)
			except (KeyError, RuntimeError):
				if sys.exc_type == KeyError:
					warn("in group \"%s\": no such group \"%s\"" % (groupName, list))
				if sys.exc_type == RuntimeError:
					err("runtime error: possible loop in configuration file")
		return out
	
	def getParam(self, param, group=None):
		"""Return the value for param
	
		If group is specified, return the groups local value. If the
		group has no local value or group=None or group does not exist,
		return the global value.

		If there's no value for param, return the empty string."""

		def getGlobalParam(param):
			try: return self["settings"][param]
			except KeyError: return ''

		try:
			val = self.getGroup(group)[param]
			if val == '': return getGlobalParam(param)
			else: return val
		except KeyError:
			return getGlobalParam(param)
